#! /usr/bin/env python3

# ######################################################################## #
# File:     tests/test_time_types.py
#
# Purpose:  Unit-test for `asynkio.time.Duration` and
#           `asynkio.time.Interval`.
#
# Created:  25th July 2025
# Updated:  4th August 2025
#
# Copyright (c) Matthew Wilson, Synesis Information Systems Pty Ltd
# All rights reserved
#
# ######################################################################## #


import math
from asynkio.time import (
    Duration,
    Instant,
)


# ######################################
# Duration

def test_Duration_DURATION_0():

    duration = Duration(0, 0)

    assert 0 == duration.as_nanos()
    assert 0 == duration.as_micros()
    assert 0 == duration.as_millis()
    assert 0 == duration.as_secs()
    assert 0.0 == duration.as_secs_f()

    assert 0 == duration.subsec_nanos()
    assert 0 == duration.subsec_micros()
    assert 0 == duration.subsec_millis()

    assert "0s" == str(duration)


def test_Duration_DURATION_1_SECOND():

    duration = Duration.from_secs(1)

    assert 1_000_000_000 == duration.as_nanos()
    assert 1_000_000 == duration.as_micros()
    assert 1_000 == duration.as_millis()
    assert 1 == duration.as_secs()
    assert math.isclose(1.0, duration.as_secs_f())

    assert 0 == duration.subsec_nanos()
    assert 0 == duration.subsec_micros()
    assert 0 == duration.subsec_millis()

    assert "1s" == str(duration)


def test_Duration_DURATION_123_MILLISECONDS():

    duration = Duration.from_millis(123)

    assert 123_000_000 == duration.as_nanos()
    assert 123_000 == duration.as_micros()
    assert 123 == duration.as_millis()
    assert 0 == duration.as_secs()
    assert math.isclose(0.123, duration.as_secs_f())

    assert 123_000_000 == duration.subsec_nanos()
    assert 123_000 == duration.subsec_micros()
    assert 123 == duration.subsec_millis()

    assert "123ms" == str(duration)


def test_Duration_DURATION_123_456_789_NANOSECONDS():

    duration = Duration.from_nanos(123_456_789)

    assert 123_456_789 == duration.as_nanos()
    assert 123_456 == duration.as_micros()
    assert 123 == duration.as_millis()
    assert 0 == duration.as_secs()
    assert math.isclose(0.123_456_789, duration.as_secs_f())

    assert 123_456_789 == duration.subsec_nanos()
    assert 123_456 == duration.subsec_micros()
    assert 123 == duration.subsec_millis()

    assert "123.4ms" == str(duration)


def test_Duration_REPR():

    assert "<asynkio.time.types.Duration: _duration=0>" == repr(Duration.from_nanos(0))


def test_Duration_STRINGS():

    assert "9ns" == str(Duration.from_nanos(9))
    assert "89ns" == str(Duration.from_nanos(89))
    assert "789ns" == str(Duration.from_nanos(789))

    assert "6.789µs" == str(Duration.from_nanos(6_789))
    assert "56.78µs" == str(Duration.from_nanos(56_789))
    assert "456.7µs" == str(Duration.from_nanos(456_789))

    assert "3.456ms" == str(Duration.from_nanos(3_456_789))
    assert "23.45ms" == str(Duration.from_nanos(23_456_789))
    assert "123.4ms" == str(Duration.from_nanos(123_456_789))

    assert "9.123s" == str(Duration.from_nanos(9_123_456_789))
    assert "89.12s" == str(Duration.from_nanos(89_123_456_789))
    assert "789.1s" == str(Duration.from_nanos(789_123_456_789))


    assert "9ns" == str(Duration.from_nanos(9))
    assert "80ns" == str(Duration.from_nanos(80))
    assert "700ns" == str(Duration.from_nanos(700))

    assert "6µs" == str(Duration.from_nanos(6_000))
    assert "50µs" == str(Duration.from_nanos(50_000))
    assert "400µs" == str(Duration.from_nanos(400_000))

    assert "3ms" == str(Duration.from_nanos(3_000_000))
    assert "20ms" == str(Duration.from_nanos(20_000_000))
    assert "100ms" == str(Duration.from_nanos(100_000_000))

    assert "9s" == str(Duration.from_nanos(9_000_000_000))
    assert "10s" == str(Duration.from_nanos(10_000_000_000))
    assert "200s" == str(Duration.from_nanos(200_000_000_000))
    assert "3000s" == str(Duration.from_nanos(3_000_000_000_000))
    assert "40000s" == str(Duration.from_nanos(40_000_000_000_000))
    assert "500000s" == str(Duration.from_nanos(500_000_000_000_000))
    assert "6000000s" == str(Duration.from_nanos(6_000_000_000_000_000))
    assert "70000000s" == str(Duration.from_nanos(70_000_000_000_000_000))

    assert "11.11s" == str(Duration.from_nanos(11_111_111_111))
    assert "222.2s" == str(Duration.from_nanos(222_222_222_222))
    assert "3333s" == str(Duration.from_nanos(3_333_333_333_333))
    assert "44444s" == str(Duration.from_nanos(44_444_444_444_444))
    assert "555555s" == str(Duration.from_nanos(555_555_555_555_555))
    assert "6666666s" == str(Duration.from_nanos(6_666_666_666_666_666))
    assert "77777777s" == str(Duration.from_nanos(77_777_777_777_777_777))

    assert "-11.11s" == str(Duration.from_nanos(-11_111_111_111))
    assert "-222.2s" == str(Duration.from_nanos(-222_222_222_222))
    assert "-3333s" == str(Duration.from_nanos(-3_333_333_333_333))
    assert "-44444s" == str(Duration.from_nanos(-44_444_444_444_444))
    assert "-555555s" == str(Duration.from_nanos(-555_555_555_555_555))
    assert "-6666666s" == str(Duration.from_nanos(-6_666_666_666_666_666))
    assert "-77777777s" == str(Duration.from_nanos(-77_777_777_777_777_777))


def test_Duration_NEGATIVE_VALUES_STRINGS():

    assert "-9ns" == str(Duration.from_nanos(-9))
    assert "-89ns" == str(Duration.from_nanos(-89))
    assert "-789ns" == str(Duration.from_nanos(-789))

    assert "-6.789µs" == str(Duration.from_nanos(-6_789))
    assert "-56.78µs" == str(Duration.from_nanos(-56_789))
    assert "-456.7µs" == str(Duration.from_nanos(-456_789))

    assert "-3.456ms" == str(Duration.from_nanos(-3_456_789))
    assert "-23.45ms" == str(Duration.from_nanos(-23_456_789))
    assert "-123.4ms" == str(Duration.from_nanos(-123_456_789))

    assert "-9.123s" == str(Duration.from_nanos(-9_123_456_789))


    assert "-9ns" == str(Duration.from_nanos(-9))
    assert "-80ns" == str(Duration.from_nanos(-80))
    assert "-700ns" == str(Duration.from_nanos(-700))

    assert "-6µs" == str(Duration.from_nanos(-6_000))
    assert "-50µs" == str(Duration.from_nanos(-50_000))
    assert "-400µs" == str(Duration.from_nanos(-400_000))

    assert "-3ms" == str(Duration.from_nanos(-3_000_000))
    assert "-20ms" == str(Duration.from_nanos(-20_000_000))
    assert "-100ms" == str(Duration.from_nanos(-100_000_000))

    assert "-9s" == str(Duration.from_nanos(-9_000_000_000))
    assert "-10s" == str(Duration.from_nanos(-10_000_000_000))
    assert "-200s" == str(Duration.from_nanos(-200_000_000_000))
    assert "-3000s" == str(Duration.from_nanos(-3_000_000_000_000))
    assert "-40000s" == str(Duration.from_nanos(-40_000_000_000_000))


def test_Duration_OBSERVED_EDGE_CASES():

    assert "999.7ms" == str(Duration.from_nanos(999_772_000))
    assert "999.8ms" == str(Duration.from_nanos(999_800_000))
    assert "999.9ms" == str(Duration.from_nanos(999_974_000))

    assert "-999.7ms" == str(Duration.from_nanos(-999_772_000))
    assert "-999.8ms" == str(Duration.from_nanos(-999_800_000))
    assert "-999.9ms" == str(Duration.from_nanos(-999_974_000))


def test_Duration_WITH_PLUS_SIGN():

    assert "999.7ms" == f"{Duration.from_nanos(999_772_000)}"
    assert "999.8ms" == f"{Duration.from_nanos(999_800_000)}"
    assert "999.9ms" == f"{Duration.from_nanos(999_974_000)}"

    assert "-999.7ms" == f"{Duration.from_nanos(-999_772_000)}"
    assert "-999.8ms" == f"{Duration.from_nanos(-999_800_000)}"
    assert "-999.9ms" == f"{Duration.from_nanos(-999_974_000)}"

    assert "999.7ms" == f"{Duration.from_nanos(999_772_000):}"
    assert "999.8ms" == f"{Duration.from_nanos(999_800_000):}"
    assert "999.9ms" == f"{Duration.from_nanos(999_974_000):}"

    assert "+999.7ms" == f"{Duration.from_nanos(999_772_000):+}"
    assert "+999.8ms" == f"{Duration.from_nanos(999_800_000):+}"
    assert "+999.9ms" == f"{Duration.from_nanos(999_974_000):+}"


def test_Duration_ADD():

    d_100 = Duration.from_nanos(100)

    assert Duration.from_nanos(100) == (d_100 + Duration.from_nanos(0))
    assert Duration.from_nanos(50) == (d_100 + Duration.from_nanos(-50))
    assert Duration.from_nanos(150) == (d_100 + Duration.from_nanos(50))


def test_Duration_SUB():

    d_100 = Duration.from_nanos(100)

    assert Duration.from_nanos(100) == (d_100 - Duration.from_nanos(0))
    assert Duration.from_nanos(50) == (d_100 - Duration.from_nanos(50))
    assert Duration.from_nanos(150) == (d_100 - Duration.from_nanos(-50))


def test_Duration_DIV():

    d_100 = Duration.from_nanos(100)

    assert Duration.from_nanos(100) == (d_100 / 1)
    assert Duration.from_nanos(50) == (d_100 / 2)
    assert Duration.from_nanos(50) == (d_100 / 2.0)
    assert 1 == d_100 / Duration.from_nanos(100)


def test_Duration_MUL():

    d_100 = Duration.from_nanos(100)

    assert Duration.from_nanos(100) == (d_100 * 1)
    assert Duration.from_nanos(50) == (d_100 * 0.5)
    assert Duration.from_nanos(200) == (d_100 * 2.0)


# ######################################
# Instant

def test_Instant_CREATE():

    instant_1 = Instant.now()
    instant_2 = Instant.now()

    assert not (instant_1 > instant_2)
    assert instant_1 <= instant_2

    assert not (instant_2 < instant_1)
    assert instant_2 >= instant_1

    duration_1_2 = instant_2 - instant_1

    max_duration_delta = 5

    assert duration_1_2.as_micros() < max_duration_delta, f"{duration_1_2.as_micros()} should be < {max_duration_delta}"


    instant_3 = instant_1 + Duration.from_micros(123)

    duration_1_3 = instant_3 - instant_1

    assert 123_000 == duration_1_3.as_nanos()


def test_Intant_REPR():

    assert "<asynkio.time.types.Instant: _t=0>" == repr(Instant(0))
    assert "<asynkio.time.types.Instant: _t=123>" == repr(Instant(123))
    assert "<asynkio.time.types.Instant: _t=123456>" == repr(Instant(123_456))
    assert "<asynkio.time.types.Instant: _t=123456789>" == repr(Instant(123_456_789))
    assert "<asynkio.time.types.Instant: _t=980123456789>" == repr(Instant(980_123_456_789))


def test_Intant_STR():

    instant_0 = Instant(0)

    assert "1970-01-01T00:00:00.000000Z" == str(instant_0)

    instant_x = Instant(1_754_271_065_290_980_000)

    assert "2025-08-04T01:31:05.290980Z" == str(instant_x)


def test_Intant_FMT():

    instant_0 = Instant(0)

    assert "1970-01-01T00:00:00.000000Z" == format(instant_0, '')

    assert "0" == format(instant_0, 'd')


    instant_x = Instant(1_754_271_065_290_980_000)

    assert "2025-08-04T01:31:05.290980Z" == format(instant_x, '')

    assert "1754271065290980000" == format(instant_x, 'd')
    assert "141303303650632435240" == format(instant_x, 'o')
    assert "18586c3d466a3aa0" == format(instant_x, 'x')
    assert "18586C3D466A3AA0" == format(instant_x, 'X')

    assert "+1754271065290980000" == format(instant_x, '+d')
    assert "+141303303650632435240" == format(instant_x, '+o')
    assert "+18586c3d466a3aa0" == format(instant_x, '+x')
    assert "+18586C3D466A3AA0" == format(instant_x, '+X')

    assert "1754271065290980000" == format(instant_x, '#d')
    assert "0o141303303650632435240" == format(instant_x, '#o')
    assert "0x18586c3d466a3aa0" == format(instant_x, '#x')
    assert "0X18586C3D466A3AA0" == format(instant_x, '#X')


def test_Intant_INT():

    assert 0 == Instant(0).__int__()
    assert 123 == Instant(123).__int__()





from datetime import (
    datetime,
    timezone,
)
import time
from typing import Self


class Duration:
    """
    Represents a span of time.
    """

    __slots__ = (
        "_duration",
    )

    def __init__(
        self,
        from_ns,
        to_ns,
    ):
        """
        Creates a new instance from the number of nanoseconds represented by
        `to_ns - from_ns`.
        """

        self._duration = to_ns - from_ns

    @staticmethod
    def from_nanos(t_ns : int) -> Self:
        """
        Creates a new instance from the specified number of nanoseconds.
        """

        return Duration(0, t_ns)

    @staticmethod
    def from_micros(t_us : int) -> Self:
        """
        Creates a new instance from the specified number of microseconds.
        """

        return Duration(0, t_us * 1_000)

    @staticmethod
    def from_millis(t_ms : int) -> Self:
        """
        Creates a new instance from the specified number of milliseconds.
        """

        return Duration(0, t_ms * 1_000_000)

    @staticmethod
    def from_secs(t_s : int) -> Self:
        """
        Creates a new instance from the specified number of seconds.
        """

        return Duration(0, t_s * 1_000_000_000)

    def as_nanos(self) -> int:
        """
        Total number of whole nanoseconds contained by this instance.
        """

        return self._duration

    def as_micros(self) -> int:
        """
        Total number of whole microseconds contained by this instance.
        """

        return self._duration // 1_000

    def as_millis(self) -> int:
        """
        Total number of whole milliseconds contained by this instance.
        """

        return self._duration // 1_000_000

    def as_secs(self) -> int:
        """
        Total number of whole seconds contained by this instance.
        """

        return self._duration // 1_000_000_000

    def as_secs_f(self) -> float:
        """
        The fractional number of seconds contained by this instance.
        """

        return self._duration / 1_000_000_000.0

    def subsec_nanos(self) -> int:
        """
        The fractional part of this instance, in nanoseconds.
        """

        return self._duration % 1_000_000_000

    def subsec_micros(self) -> int:
        """
        The fractional part of this instance, in microseconds.
        """

        return (self._duration % 1_000_000_000) // 1_000

    def subsec_millis(self) -> int:
        """
        The fractional part of this instance, in milliseconds.
        """

        return (self._duration % 1_000_000_000) // 1_000_000

    @staticmethod
    def _scale_index(n : int) -> tuple[int, int]:

        _SCALES = [
            1,
            10,
            100,
            1_000,
            10_000,
            100_000,
            1_000_000,
            10_000_000,
            100_000_000,
            1_000_000_000,
            10_000_000_000,
            100_000_000_000,
        ]

        assert n > 0
        assert len(_SCALES) == 12

        if n >= 100_000_000_000:

            return (11, _SCALES[11])

        l = 0
        h = 11

        count = 0

        while l <= h:

            count += 1

            assert count < 5, f"too many loops while trying to scale {n}"

            m = (h + l) // 2

            b = _SCALES[m]

            if n == b:

                return (m, b)

            if n < b:

                h = m

                continue
            else:
                assert n > b

                if n < b * 10:

                    return (m, b)
                else:

                    l = m

        return (11, _SCALES[11])

    @staticmethod
    def duration_to_string(
        duration : Self | int,
        format_spec : str = '',
    ) -> str:

        v = int(duration)

        if v < 0:

            v = -v
            sign = '-'
        else:

            if '+' in format_spec:

                sign = '+'
            else:

                sign = ''

        if v == 0:

            return "0s"

        oom, divisor = Duration._scale_index(v)

        suffixes = [
            'ns',
            'µs',
            'ms',
            's',
        ]
        suffix = suffixes[oom // 3]

        def fmt(sign, whole, frac, suffix):

            if frac == 0:

                return f"{sign}{whole}{suffix}"
            else:

                if whole > 999:

                    return f"{sign}{whole}{suffix}"

                if whole > 99:

                    return f"{sign}{whole}.{frac}{suffix}"

                if whole > 9:

                    return f"{sign}{whole}.{frac:02d}{suffix}"

                return f"{sign}{whole}.{frac}{suffix}"

        if oom < 3:

            return f"{sign}{v}{suffix}"
        else:

            divisor_0 = divisor // 1_000

            i = oom % 3

            if i == 0:

                divisor_1 = 1_000
            elif i == 1:

                divisor_1 = 100
            else:

                divisor_1 = 10


            v //= divisor_0

            whole = v // divisor_1
            frac = v - (whole * divisor_1)


            return fmt(sign, whole, frac, suffix)

    def __eq__(self, rhs : Self | float | int) -> bool:

        if isinstance(rhs, Duration):

            return self._duration == rhs._duration

        if isinstance(rhs, (float, int)):

            return self._duration == int(rhs)

        return NotImplemented

    def __format__(self, format_spec) -> str:

        return self.duration_to_string(
            self._duration,
            format_spec=format_spec,
        )

    def __str__(self) -> str:

        return self.duration_to_string(
            self._duration,
            format_spec='',
        )

    def __repr__(self) -> str:

        return f"<{self.__module__}.{self.__class__.__name__}: _duration={self._duration}>"

    def __int__(self) -> int:
        """
        Converts an instance into an integer, representing the number of
        nanoseconds in the duration.
        """

        return self._duration

    def __add__(self, rhs : Self) -> Self:

        if isinstance(rhs, Duration):

            return Duration.from_nanos(self._duration + rhs._duration)

        return NotImplemented

    def __sub__(self, rhs : Self) -> Self:

        if isinstance(rhs, Duration):

            return Duration.from_nanos(self._duration - rhs._duration)

        return NotImplemented

    def __mul__(self, rhs : float | int) -> Self:

        if isinstance(rhs, float):

            return Duration.from_nanos(int(self._duration * rhs))

        if isinstance(rhs, int):

            return Duration.from_nanos(self._duration * rhs)

        return NotImplemented

    def __truediv__(self, rhs : Self | float | int) -> Self:

        if isinstance(rhs, (Duration, int)):

            return Duration.from_nanos(self._duration / int(rhs))

        if isinstance(rhs, float):

            return Duration.from_nanos(int(self._duration / rhs))

        return NotImplemented


class Instant:
    """
    Represents a moment in time.
    """

    __slots__ = (
        "_t",
    )

    def __init__(self, t_ns):
        """
        Initialises with the given time instant (specified in nanoseconds).
        """

        self._t = t_ns

    @staticmethod
    def _t_ns_to_d_utc(t_ns : int) -> datetime:
        """
        Convert the number of nanoseconds since epoch into a datetime
        instance assuming UTC.
        """

        return datetime.fromtimestamp(t_ns / 1_000_000_000.0, tz=timezone.utc)

    @staticmethod
    def now() -> Self:
        """
        Initialises with the current time instant.
        """

        t_now_ns = time.time_ns()

        return Instant(t_now_ns)

    @staticmethod
    def instant_to_string(
        instant : Self | int,
        format_spec : str = '',
    ) -> str:

        typed = None

        def set_type_or_raise(typed, t, c, b):

            # local as_type

            if typed is not None:

                raise ValueError(f"cannot specify type `{t}` as type already specified as `{typed[0]}`")

            else:

                return (t, c, b)

        show_plus = False
        show_base = False

        for c in format_spec:

            print(f"c={c}")

            if c == 'd':

                typed = set_type_or_raise(typed, int, c, '')

            if c == 'o':

                typed = set_type_or_raise(typed, int, c, '0o')

            if c == 'x':

                typed = set_type_or_raise(typed, int, c, '0x')

            if c == 'X':

                typed = set_type_or_raise(typed, int, c, '0X')

            if c == '+':

                show_plus = True

            if c == '#':

                show_base = True

        if typed:

            if typed[0] == int:

                fmt = ''

                if show_plus:

                    fmt += '+'

                if show_base:

                    fmt += '#'

                fmt += typed[1]

                return format(instant, fmt)
            else:

                return NotImplemented
        else:

            d = Instant._t_ns_to_d_utc(instant)

            return d.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    def __format__(self, format_spec) -> str:

        return Instant.instant_to_string(
            self._t,
            format_spec=format_spec,
        )

    def __str__(self) -> str:

        return Instant.instant_to_string(
            self._t,
            format_spec='',
        )

    def __repr__(self) -> str:

        return f"<{self.__module__}.{self.__class__.__name__}: _t={self._t}>"

    def __int__(self) -> int:
        """
        Converts an instance into an integer, representing the number of
        nanoseconds since the epoch.
        """

        return self._t

    def __lt__(self, rhs) -> bool:
        """
        Determines whether the called instance is less-than the `rhs`
        instance of `Instant`.
        """

        if isinstance(rhs, Instant):

            return self._t < rhs._t

        raise NotImplemented

    def __le__(self, rhs) -> bool:
        """
        Determines whether the called instance is less-than-or-equal-to the
        `rhs` instance of `Instant`.
        """

        if isinstance(rhs, Instant):

            return self._t <= rhs._t

        raise NotImplemented

    def __sub__(self, rhs) -> Duration | Self:
        """
        Subtracts the `rhs` - which may be an instance of `Instant` or an
        instance of `Duration` - from the called instance.
        """

        if isinstance(rhs, Instant):

            return Duration(rhs._t, self._t)

        if isinstance(rhs, Duration):

            return Instant(self._t - rhs.as_nanos())

        return NotImplemented

    def __add__(self, rhs) -> Self:
        """
        Adds the `rhs` instance of `Duration` to the called instance.
        """

        if isinstance(rhs, Duration):

            return Instant(self._t + rhs.as_nanos())

        return NotImplemented

